"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const Cheerio = __importStar(require("cheerio"));
const XmlDom = __importStar(require("xmldom"));
const Constants_1 = require("./Constants");
const HtmlQuotations_1 = require("./HtmlQuotations");
const Regexp_1 = require("./Regexp");
const Utils_1 = require("./Utils");
const xmlDomParser = new XmlDom.DOMParser({ errorHandler: { warning: () => { }, error: () => { }, fatalError: (error) => { throw error; } } });
const xmlDomSerializer = new XmlDom.XMLSerializer();
function extractFromPlain(messageBody, options) {
    if (!messageBody || !messageBody.trim())
        return { body: messageBody, didFindQuote: false };
    const { maxLinesCount } = {
        maxLinesCount: Constants_1.DefaultMaxLinesCount,
        ...options
    };
    const delimiter = Utils_1.findDelimiter(messageBody);
    messageBody = preprocess(messageBody, delimiter);
    const lines = Utils_1.splitLines(messageBody).slice(0, maxLinesCount);
    const markers = markMessageLines(lines);
    const { wereLinesDeleted, lastMessageLines } = processMarkedLines(lines, markers);
    messageBody = lastMessageLines.join(delimiter);
    messageBody = postProcess(messageBody);
    return { body: messageBody, didFindQuote: wereLinesDeleted };
}
exports.extractFromPlain = extractFromPlain;
function extractFromHtml(messageBody, options = {}) {
    if (!messageBody || !messageBody.trim())
        return { body: messageBody, didFindQuote: false };
    messageBody = messageBody.replace(/\r\n/g, " ").replace(/\n/g, " ");
    messageBody = Utils_1.normalizeHtmlDocument(messageBody);
    const document = loadHtmlAndFix(messageBody);
    const xmlDocument = xmlDomParser.parseFromString(document.xml());
    if (xmlDocument.lastChild && xmlDocument.lastChild.nodeValue)
        xmlDocument.removeChild(xmlDocument.lastChild);
    const result1 = extractQuotationUsingCheckpoints(xmlDocument, messageBody, options);
    if ((result1.didFindQuote && !result1.isTooLong) || result1.isForwardedMessage)
        return result1;
    let wasQuotationRemoved = cutQuotation(xmlDocument);
    if (!wasQuotationRemoved)
        return { body: messageBody, didFindQuote: false };
    if (result1.isTooLong) {
        const result2 = extractQuotationUsingCheckpoints(xmlDocument, messageBody, options);
        if (result2.didFindQuote && !result2.isTooLong)
            return result2;
    }
    else if (options.skipSimpleQuoteRemoval) {
        return {
            body: messageBody,
            didFindQuote: false,
            didUseCheckpoints: true
        };
    }
    return {
        body: xmlDomSerializer.serializeToString(xmlDocument, true),
        didFindQuote: true,
        didUseCheckpoints: false
    };
}
exports.extractFromHtml = extractFromHtml;
function extractQuotationUsingCheckpoints(xmlDocument, messageBody, options) {
    const xmlDocumentCopy = xmlDocument.cloneNode(true);
    const { nodeLimit, maxLinesCount } = {
        nodeLimit: Constants_1.DefaultNodeLimit,
        maxLinesCount: Constants_1.DefaultMaxLinesCount,
        ...options
    };
    let numberOfCheckpoints = HtmlQuotations_1.addCheckpoint(xmlDocumentCopy, xmlDocumentCopy, { nodeLimit });
    if (numberOfCheckpoints >= nodeLimit)
        return { body: messageBody, didFindQuote: false, isTooLong: true };
    let extractQuoteHtml = extractQuoteHtmlViaMarkers(numberOfCheckpoints, xmlDocumentCopy, { ignoreBlockTags: false, maxLinesCount });
    if (extractQuoteHtml.error)
        return { body: messageBody, didFindQuote: false, isTooLong: true };
    if (extractQuoteHtml.isForwardedMessage)
        return {
            body: xmlDomSerializer.serializeToString(xmlDocument, true),
            didFindQuote: false,
            didUseCheckpoints: true,
            isForwardedMessage: true
        };
    if (!extractQuoteHtml.quoteWasFound)
        extractQuoteHtml = extractQuoteHtmlViaMarkers(numberOfCheckpoints, xmlDocumentCopy, { ignoreBlockTags: true, maxLinesCount });
    if (!extractQuoteHtml.quoteWasFound)
        return { body: null, didFindQuote: false, isTooLong: false };
    HtmlQuotations_1.deleteQuotationTags(xmlDocumentCopy, xmlDocument, extractQuoteHtml.quotationCheckpoints, { nodeLimit });
    if (xmlDocument.nodeType === Constants_1.NodeTypes.DOCUMENT_NODE && !xmlDocument.documentElement)
        xmlDocument.documentElement = xmlDocument.childNodes[0];
    cutQuotation(xmlDocument, { onlyRemoveEmptyBlocks: true });
    return {
        body: xmlDomSerializer.serializeToString(xmlDocument, true),
        didFindQuote: true,
        didUseCheckpoints: true
    };
}
function cutQuotation(xmlDocument, options) {
    return HtmlQuotations_1.cutGmailQuote(xmlDocument, options)
        || HtmlQuotations_1.cutZimbraQuote(xmlDocument, options)
        || HtmlQuotations_1.cutBlockquote(xmlDocument, options)
        || HtmlQuotations_1.cutMicrosoftQuote(xmlDocument, options)
        || HtmlQuotations_1.cutById(xmlDocument, options);
}
function extractQuoteHtmlViaMarkers(numberOfCheckpoints, xmlDocument, options) {
    const { ignoreBlockTags, maxLinesCount } = {
        ignoreBlockTags: false,
        maxLinesCount: Constants_1.DefaultMaxLinesCount,
        ...options
    };
    const messagePlainText = preprocess(Utils_1.elementToText(xmlDocument, { ignoreBlockTags }), "\n", Constants_1.ContentTypeTextHtml);
    let lines = Utils_1.splitLines(messagePlainText);
    if (lines.length > maxLinesCount)
        return { error: 'Message too big' };
    const lineCheckpoints = lines.map(line => {
        const match = line.match(new RegExp(Regexp_1.CheckPointRegexp.source, "g"));
        return match
            ? match.map(matchPart => parseInt(matchPart.slice(4, -4), 10))
            : new Array();
    });
    lines = lines.map(line => line.replace(new RegExp(Regexp_1.CheckPointRegexp.source, "g"), ""));
    const markers = markMessageLines(lines);
    const { wereLinesDeleted, firstDeletedLine, lastDeletedLine, isForwardedMessage } = processMarkedLines(lines, markers);
    const quotationCheckpoints = new Array(numberOfCheckpoints);
    if (wereLinesDeleted)
        for (let index = firstDeletedLine; index <= lastDeletedLine; index++)
            for (let checkpoint of lineCheckpoints[index])
                quotationCheckpoints[checkpoint] = true;
    return { quoteWasFound: wereLinesDeleted, quotationCheckpoints, isForwardedMessage };
}
function preprocess(messageBody, delimiter, contentType = Constants_1.ContentTypeTextPlain) {
    messageBody = messageBody.replace(new RegExp(Regexp_1.LinkRegexp.source, "g"), (match, link, offset, str) => {
        const newLineIndex = str.substring(offset).indexOf("\n");
        return str[newLineIndex + 1] === ">" ? match : `@@${link}@@`;
    });
    if (contentType !== Constants_1.ContentTypeTextPlain)
        return messageBody;
    messageBody = messageBody.replace(new RegExp(Regexp_1.OnDateSomebodyWroteRegexp.source, "g"), (match, ...args) => {
        const offset = args.filter(a => isFinite(a))[0];
        const str = args[args.length - 1];
        return offset > 0 && str[offset - 1] !== "\n"
            ? delimiter + match
            : match;
    });
    return messageBody;
}
exports.preprocess = preprocess;
function markMessageLines(lines) {
    const markers = new Array(lines.length);
    let index = 0;
    while (index < lines.length) {
        const line = lines[index];
        if (!line || !line.trim()) {
            markers[index] = "e";
        }
        else if (Utils_1.matchStart(line, Regexp_1.QuotePatternRegexp)) {
            markers[index] = "m";
        }
        else if (Utils_1.isStartOfForwardedMessage(line)) {
            markers[index] = "f";
        }
        else {
            const splitterMatch = isSplitter(lines.slice(index, index + Constants_1.SplitterMaxLines).join("\n"));
            if (!splitterMatch) {
                markers[index] = "t";
            }
            else {
                const splitterLines = Utils_1.splitLines(splitterMatch[0]);
                for (let splitterIndex = 0; splitterIndex < splitterLines.length; splitterIndex++)
                    markers[index + splitterIndex] = "s";
                index += splitterLines.length - 1;
            }
        }
        index++;
    }
    return markers.join("");
}
exports.markMessageLines = markMessageLines;
function processMarkedLines(lines, markers) {
    const result = {
        isForwardedMessage: false,
        firstDeletedLine: -1,
        lastDeletedLine: -1,
        lastMessageLines: lines,
        wereLinesDeleted: false
    };
    if (markers.indexOf("s") < 0 && !/(me*){3}/.exec(markers))
        markers = markers.replace(/m/g, "t");
    if (Utils_1.matchStart(markers, /[te]*f/))
        return { ...result, isForwardedMessage: true };
    let inlineReplyMatch;
    const inlineReplyRegexp = /me*(te*)+m/g;
    while (inlineReplyMatch = inlineReplyRegexp.exec(markers)) {
        const links = lines[inlineReplyMatch.index].match(Regexp_1.ParenthesisLinkRegexp)
            || Utils_1.matchStart(lines[inlineReplyMatch.index + 1].trim(), Regexp_1.ParenthesisLinkRegexp);
        if (!links)
            return result;
        inlineReplyRegexp.lastIndex--;
    }
    let quotation = markers.match("(e*)(se*)+((t|f)+e*)+");
    if (quotation) {
        result.wereLinesDeleted = true;
        result.firstDeletedLine = quotation.index;
        result.lastDeletedLine = lines.length - 1;
        result.lastMessageLines = lines.slice(0, quotation.index);
        return result;
    }
    quotation = markers.match(Regexp_1.QuotationAfterSplitterRegexp)
        || markers.match(Regexp_1.QuotationBlockRegexp)
        || markers.match(Regexp_1.EmptyQuotationAfterSplitterRegexp)
        || markers.match(Regexp_1.EmptyQuotationBlockRegexp);
    if (quotation) {
        const firstGroupStart = quotation.index + quotation[0].indexOf(quotation[1]);
        const firstGroupEnd = firstGroupStart + quotation[1].length;
        result.wereLinesDeleted = true;
        result.firstDeletedLine = firstGroupStart;
        result.lastDeletedLine = firstGroupEnd - 1;
        result.lastMessageLines = lines.slice(0, firstGroupStart).concat(lines.slice(firstGroupEnd));
        return result;
    }
    return result;
}
exports.processMarkedLines = processMarkedLines;
function postProcess(messageBody) {
    return messageBody.replace(new RegExp(Regexp_1.NormalizedLinkRegexp.source, "g"), "<$1>").trim();
}
function isSplitter(src) {
    for (const pattern of Regexp_1.SplitterRegexps) {
        var match = Utils_1.matchStart(src, pattern);
        if (match)
            return match;
    }
}
function loadHtmlAndFix(src) {
    const document = Cheerio.load(src);
    document.root().find("*").contents().filter((n, e) => e.type === "directive" && e.name[0] === "!").remove();
    return document;
}
//# sourceMappingURL=Quotations.js.map